#include <cfloat>
#include <error.h>
#include <matutil.h>
#include <raices.h>


/*
*  Funciones auxiliares de los algoritmos de bsqueda de races:
*/

/*
 *  METODO DE BISECCIN
 *  ===================
 */

/*************** bisecIni *************************************
 * - Inicializa el mtodo de biseccin
 * - Toma como argumentos un puntero a la funcin f y los extremos 
 * - del intervalo
 * - Comprueba que la raz est acotada (fa*fb < 0)
 * - devuelve EXITO si la raz est acotada y un cdigo de error en 
 * - otro caso
 */
int bisecIni(func F, double a, double b)
{

  // Calcula las imgenes de los extremos del intervalo

  double fa, fb;

  fa = F(a);
  // la funcin debe ser continua
  if (!finite(fa))
    return CEFUNC;

  fb = F(b);  
  // la funcin debe ser continua
  if (!finite(fb))
    return CEFUNC;

  // Devuelve CNOACOTADO si la funcin no cambia de signo
  if( fa*fb >0)    
    return CNOACOTADO;
    
  return EXITO;
}

/*************** bisecIter *************************************
 * Ejecuta una iteracin del mtodo de biseccin. 
 * - Si se encuentra la raz exacta la funcin devuelve EXITO 
 * - Si la funcin es discontinua devuelve un cdigo de error
 * - En otro caso devuelve la aproximacin a la raz en el argumento
 * y el cdigo ITERANDO
 */
int bisecIter(func F, double& a, double& b, double& raiz)
{

  // Calcula las imgenes de los extremos del intervalo
  double fa, fb, f_raiz;
  fa = F(a);
  fb = F(b);

  // Dividimos el intervalo y calculamos la imagen del punto medio
  raiz = (a + b)/2.0;
  f_raiz = F(raiz);
  // la funcin debe ser continua
  if (!finite(f_raiz))
    return CEFUNC;

  // Comprobamos si hemos dado con la  raz exacta
  if(f_raiz == 0.0) 
    return EXITO;

  // O bien nos quedamos con la mitad del intervalo
  // que encierra la raz
  
  if ( fa * f_raiz < 0.0 ) 
    b = raiz;
  else 
    a = raiz;

  return ITERANDO;
}


/*
 *  METODO DE REGULA FALSI
 *  ======================
 */


/*************** rfalsiIni *************************************
 * - Inicializa el mtodo de regula Falsi
 * - Toma como argumentos un puntero a la funcin f y los extremos 
 * - del intervalo
 * - Comprueba que la raz est acotada (fa*fb < 0)
 * - devuelve EXITO si la raz est acotada y un cdigo de error en 
 * - otro caso
 */
int rfalsiIni(func F, double a, double b)
{

  // Calcula las imgenes de los extremos del intervalo

  double fa, fb;

  fa = F(a);
  // la funcin debe ser continua
  if (!finite(fa))
    return CEFUNC;

  fb = F(b);  
  // la funcin debe ser continua
  if (!finite(fb))
    return CEFUNC;

  // Devuelve CNOACOTADO si la funcin no cambia de signo
  if( fa*fb >0) 	    
    return CNOACOTADO;
    
  return EXITO;
}

/*************** rfalsiIter *************************************
 * Ejecuta una iteracin del mtodo de Regula Falsi. 
 * - Si se encuentra la raz exacta la funcin devuelve EXITO 
 * - Si la funcin es discontinua devuelve un cdigo de error
 * - En otro caso devuelve la aproximacin a la raz en el argumento
 * y el cdigo ITERANDO
 */

int rfalsiIter (func F, double& a, double& b, double& raiz)
{

  // Calcula las imgenes de los extremos del intervalo
  double fa, fb, f_raiz;
  fa = F(a);
  fb = F(b);

  // Calculamos la recta que pasa por (a,fa) y (b,fb)
  if(fa == fb) return CEZERODIV;
  raiz = (fa*b-fb*a)/(fa-fb);
  f_raiz = F(raiz);
  // la funcin debe ser continua
  if (!finite(f_raiz))
    return CEFUNC;

  // Comprobamos si hemos dado con la  raz exacta
  if(f_raiz == 0.0) 
    return EXITO;

  // O bien nos quedamos con la mitad del intervalo
  // que encierra la raz
  
  if ( fa * f_raiz < 0.0 ) 
    b = raiz;
  else 
    a = raiz;

  return ITERANDO;
}



/*
 *  METODO DE NEWTON RAPHSON
 *  ========================
 */

/*************** newtonIni *************************************
 * Inicializa el mtodo de newton rapson
 * - Calcula la imagen de la raz aproximada y su derivada
 * - Comprueba que son finitas
 */
int newtonIni (func F, func dF, double& f_raiz, double& df_raiz, double& raiz)
{
  f_raiz = F(raiz);
  // la funcin debe ser continua
  if (!finite(f_raiz))
    return CEFUNC;

  df_raiz = dF(raiz);
  // la funcin debe ser continua
  if (!finite(df_raiz))
    return CEFUNC;

  return EXITO;
}


/*************** newtonIter *************************************
 * Ejecuta una iteracin del mtodo de newton rapson
 * - Si la funcin es discontinua o la derivada nula 
 * - devuelve un cdigo de error
 * - En otro caso devuelve la aproximacin a la raz en el argumento
 * y el cdigo ITERANDO
 */
int newtonIter (func F, func dF, double& f_raiz, double& df_raiz, double& raiz)
{

  f_raiz = F(raiz);
  // la funcin debe ser continua
  if (!finite(f_raiz))
    return CEFUNC;

  df_raiz = dF(raiz);
  // la funcin debe ser continua
  if (!finite(df_raiz))
    return CEFUNC;

  // La derivada es cero, newton-raphson falla en esta
  // iteracin
  
  if (df_raiz == 0.0)
    return CEZERODIV;

  // damos un paso por el mtodo de Newton-Raphson
  // La aproximacin es:   raiz = raiz - (f_raiz / df_raiz),
  // correspondiente a quedarnos a primer orden en la expansin de
  // Taylor de f

  raiz = raiz - (f_raiz/df_raiz);

  return ITERANDO;
}











